/*
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 */

#define DRIVER_NAME "MGACON"
#define DBG2(s...) if(debug>=2)printk( KERN_DEBUG DRIVER_NAME ": " s)

#define VT_BUF_HAVE_MEMSETW
#define VT_BUF_HAVE_MEMCPYW
#define VT_BUF_HAVE_MEMMOVEW

#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/console.h>
//#include <linux/console_struct.h>
#include <linux/string.h>
#include <linux/kd.h>
#include <linux/slab.h>
#include <linux/vt_kern.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/io.h>

#define BLANK 0x2007

/*
 *  Interface used by the world
 */

static spinlock_t vga_lock = SPIN_LOCK_UNLOCKED;

static const char *mgacon_startup(void);
static void mgacon_init(struct vc_data *c, int init);
static void mgacon_deinit(struct vc_data *c);
static void mgacon_cursor(struct vc_data *c, int mode);
static int mgacon_switch(struct vc_data *c);
static int mgacon_blank(struct vc_data *c, int blank, int mode_switch);
static int mgacon_set_palette(struct vc_data *c, unsigned char *table);
static int mgacon_scrolldelta(struct vc_data *c, int lines);
static int mgacon_set_origin(struct vc_data *c);
static int mgacon_scroll(struct vc_data *c, int t, int b, int dir, int lines);
static u8 mgacon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse);
static void mgacon_invert_region(struct vc_data *c, u16 *p, int count);
static void mgacon_bmove(struct vc_data *c, int sy, int sx, 
			 int dy, int dx, int height, int width);
static void mgacon_clear(struct vc_data *c, int y, int x, 
			  int height, int width);

//static unsigned long mgacon_uni_pagedir[2];

struct consw mga_con;

/* Description of the hardware situation */
static unsigned long bar0;
static unsigned long bar1;
static char *mga_vram_base;		/*  */
static char *video_buf;		/*  */
static char *mga_mmio_base0;		/* misc, seq and graphics*/

static int mga_scan_lines;
static unsigned int    mga_video_num_columns;	/* Number of text columns */
static unsigned int    mga_video_num_lines;	/* Number of text lines */
static unsigned int    mga_default_font_height;	/* Height of default screen font */
static unsigned char   mga_font_is_default = 1;
static int	       mga_vesa_blanked;
static int	       mga_palette_blanked;
static int	       mga_is_gfx;
static int	       mga_video_font_height;
static unsigned int    mga_rolled_over = 0;
static char *mga_video_port_reg, *seq_port_reg;
static char *mga_video_port_val, *seq_port_val;
static char *video_misc_wr, *video_misc_rd;
static char *attrib_port;
static char *dac_reg, *dac_val;

static int debug=0;
static int	first_vc = 17;
static int	last_vc  = 20;

static struct vc_data	*mga_display_fg = NULL;

#ifdef MODULE_PARM
MODULE_PARM(first_vc, "i");
MODULE_PARM(last_vc,  "i");
MODULE_PARM(debug,  "i");
#endif

static inline void write_vga(unsigned char reg, unsigned int val)
{
	unsigned int v1, v2;

	v1 = reg + (val & 0xff00);
	v2 = reg + 1 + ((val << 8) & 0xff00);
	writew(v1, mga_video_port_reg);
    writew(v2, mga_video_port_reg);
}

#undef scr_memsetw
#undef scr_memcpyw
#undef scr_memmovew

static inline void scr_memsetw(u16 *dest, u16 val, unsigned int length) {
    for(length>>=1;length>0;length--) {
        scr_writew(val,dest);
        dest+=8;
    }
}

static inline void scr_memcpyw(u16 *dest, u16 *src, unsigned int length) {
    for(length>>=1;length>0;length--) {
        scr_writew(scr_readw(src),dest);
        dest+=8;
        src+=8;
    }
}

static inline void scr_memmovew(u16 *dest, u16 *src, unsigned int length) {
    if(dest<src)scr_memcpyw(dest,src,length); 
    else {
        length >>= 1;
        dest+=8*length;
        src+=8*length;
        for(;length>0;length--) {
            scr_writew(scr_readw(src),dest);
            dest-=8;
            src-=8;
        }
    }
}

#ifdef MODULE
static const char *mgacon_startup(void)
#else
__initfunc(static const char *mgacon_startup(void))
#endif
{
        int i;
        
	mga_video_num_lines = 25;
	mga_video_num_columns = 80;

        video_buf=kmalloc(65536, GFP_KERNEL);

        mga_vram_base=ioremap(bar1,16*64*1024);
        mga_mmio_base0=ioremap(bar0,16*1024);

        mga_video_port_reg=mga_mmio_base0+0x1c00+0x3d4;
        mga_video_port_val=mga_mmio_base0+0x1c00+0x3d5;
        seq_port_reg=mga_mmio_base0+0x1c00+0x3c4;
        seq_port_val=mga_mmio_base0+0x1c00+0x3c5;
        video_misc_wr=mga_mmio_base0+0x1c00+0x3c2;
        video_misc_rd=mga_mmio_base0+0x1c00+0x3cc;
        dac_reg=mga_mmio_base0+0x1c00+0x3c8;
        dac_val=mga_mmio_base0+0x1c00+0x3c9;
        attrib_port=mga_mmio_base0+0x1c00+0x3c0;

	for (i=0; i<16; i++) {
            	readb(mga_mmio_base0+0x1c00+0x3da);
            	writeb(i, attrib_port);
            	writeb(i, attrib_port);
	}
        writeb(0x20, attrib_port);
	
	for (i=0; i<16; i++) {
		writeb (color_table[i], dac_reg) ;
                writeb (default_red[i], dac_val) ;
                writeb (default_grn[i], dac_val) ;
                writeb (default_blu[i], dac_val) ;
        }

	mga_default_font_height = 16;
	mga_video_font_height = 16;
    mga_scan_lines = mga_video_font_height * mga_video_num_lines;

	return "MGACON";
}

static void mgacon_init(struct vc_data *c, int init)
{
	
	c->vc_can_do_color = 1;
	c->vc_complement_mask = 0x7700;
	c->vc_display_fg = &mga_display_fg;

	if (init) {
		c->vc_cols = mga_video_num_columns;
		c->vc_rows = mga_video_num_lines;
	} else {
		vc_resize(c->vc_num, mga_video_num_columns, mga_video_num_lines);
        }
	
	if (mga_display_fg == NULL)
		mga_display_fg = c;

#if 0
	unsigned long p;
	p = *c->vc_uni_pagedir_loc;
	if (c->vc_uni_pagedir_loc == &c->vc_uni_pagedir ||
	    !--c->vc_uni_pagedir_loc[1])
		con_free_unimap(c->vc_num);
	c->vc_uni_pagedir_loc = mgacon_uni_pagedir;
	mgacon_uni_pagedir[1]++;
	if (!mgacon_uni_pagedir[0] && p)
		con_set_default_unimap(c->vc_num);
#endif
}

static inline void mga_set_mem_top(struct vc_data *c)
{
	write_vga(12, (c->vc_visible_origin-c->vc_origin)/2);
}

static void mgacon_deinit(struct vc_data *c)
{
#if 0
	if (!--mgacon_uni_pagedir[1]) {
		c->vc_visible_origin = mga_vram_base;
		mga_set_mem_top(c);
		con_free_unimap(c->vc_num);
	}
	c->vc_uni_pagedir_loc = &c->vc_uni_pagedir;
	con_set_default_unimap(c->vc_num);
#endif
	if (mga_display_fg == c)
		mga_display_fg = NULL;

        kfree(video_buf);
}

static u8 mgacon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse)
{
	u8 attr = color;

	if (underline)
		attr = (attr & 0xf0) | c->vc_ulcolor;
	else if (intensity == 0)
		attr = (attr & 0xf0) | c->vc_halfcolor;
	if (reverse)
		attr = ((attr) & 0x88) | ((((attr) >> 4) | ((attr) << 4)) & 0x77);
	if (blink)
		attr ^= 0x80;
	if (intensity == 2)
		attr ^= 0x08;
	return attr;
}

static void mgacon_invert_region(struct vc_data *c, u16 *p, int count)
{
	while (count--) {
		u16 a = scr_readw(p);
			a = ((a) & 0x88ff) | (((a) & 0x7000) >> 4) | (((a) & 0x0700) << 4);
		scr_writew(a, p);
                p+=8;
	}
}

static void mgacon_set_cursor_size(int xpos, int from, int to)
{
	unsigned long flags;
	int curs, cure;
	static int lastfrom, lastto;

	if ((from == lastfrom) && (to == lastto)) return;
	lastfrom = from; lastto = to;

	spin_lock_irqsave(&vga_lock, flags);
	writeb(0x0a, mga_video_port_reg);		/* Cursor start */
	curs = readb(mga_video_port_val);
	writeb(0x0b, mga_video_port_reg);		/* Cursor end */
	cure = readb(mga_video_port_val);

	curs = (curs & 0xc0) | from;
	cure = (cure & 0xe0) | to;

        writeb(0x0a, mga_video_port_reg);		/* Cursor start */
        writeb(curs, mga_video_port_val);
        writeb(0x0b, mga_video_port_reg);		/* Cursor end */
        writeb(cure, mga_video_port_val);
	spin_unlock_irqrestore(&vga_lock, flags);
}

static void mgacon_cursor(struct vc_data *c, int mode)
{
    if (c->vc_origin != c->vc_visible_origin)
	mgacon_scrolldelta(c, 0);
    switch (mode) {
	case CM_ERASE:
	    write_vga(14, (64*1024 - 1));
	    break;

	case CM_MOVE:
	case CM_DRAW:
	    write_vga(14, c->vc_x+mga_video_num_columns*c->vc_y);

	    switch (c->vc_cursor_type & 0x0f) {
		case CUR_UNDERLINE:
			mgacon_set_cursor_size(c->vc_x, 
					mga_video_font_height - (mga_video_font_height < 10 ? 2 : 3),
					mga_video_font_height - (mga_video_font_height < 10 ? 1 : 2));
			break;
		case CUR_TWO_THIRDS:
			mgacon_set_cursor_size(c->vc_x, 
					 mga_video_font_height / 3,
					 mga_video_font_height - (mga_video_font_height < 10 ? 1 : 2));
			break;
		case CUR_LOWER_THIRD:
			mgacon_set_cursor_size(c->vc_x, 
					 (mga_video_font_height*2) / 3,
					 mga_video_font_height - (mga_video_font_height < 10 ? 1 : 2));
			break;
		case CUR_LOWER_HALF:
			mgacon_set_cursor_size(c->vc_x, 
					 mga_video_font_height / 2,
					 mga_video_font_height - (mga_video_font_height < 10 ? 1 : 2));
			break;
		case CUR_NONE:
			mgacon_set_cursor_size(c->vc_x, 31, 30);
			break;
          	default:
			mgacon_set_cursor_size(c->vc_x, 1, mga_video_font_height);
			break;
		}
	    break;
    }
}

static int mgacon_switch(struct vc_data *c)
{
	/*
	 * We need to save screen size here as it's the only way
	 * we can spot the screen has been resized and we need to
	 * set size of freshly allocated screens ourselves.
	 */
	mga_video_num_columns = c->vc_cols;
	mga_video_num_lines = c->vc_rows;
        return 1;
}

static void mga_set_palette(struct vc_data *c, unsigned char *table)
{
	int i, j ;

	for (i=j=0; i<16; i++) {
		writeb (table[i], dac_reg) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
	}
}

static int mgacon_set_palette(struct vc_data *c, unsigned char *table)
{
	if (mga_palette_blanked || !CON_IS_VISIBLE(c))
		return -EINVAL;
	mga_set_palette(c, table);
	return 0;
}

/* structure holding original VGA register settings */
static struct {
	unsigned char	SeqCtrlIndex;		/* Sequencer Index reg.   */
	unsigned char	CrtCtrlIndex;		/* CRT-Contr. Index reg.  */
	unsigned char	CrtMiscIO;		/* Miscellaneous register */
	unsigned char	HorizontalTotal;	/* CRT-Controller:00h */
	unsigned char	HorizDisplayEnd;	/* CRT-Controller:01h */
	unsigned char	StartHorizRetrace;	/* CRT-Controller:04h */
	unsigned char	EndHorizRetrace;	/* CRT-Controller:05h */
	unsigned char	Overflow;		/* CRT-Controller:07h */
	unsigned char	StartVertRetrace;	/* CRT-Controller:10h */
	unsigned char	EndVertRetrace;		/* CRT-Controller:11h */
	unsigned char	ModeControl;		/* CRT-Controller:17h */
	unsigned char	ClockingMode;		/* Seq-Controller:01h */
} mga_state;

static void mga_vesa_blank(int mode)
{
	/* save original values of VGA controller registers */
	if(!mga_vesa_blanked) {
		spin_lock_irq(&vga_lock);	
		mga_state.SeqCtrlIndex = readb(seq_port_reg);
		mga_state.CrtCtrlIndex = readb(mga_video_port_reg);
		mga_state.CrtMiscIO = readb(video_misc_rd);
		spin_unlock_irq(&vga_lock);

		writeb(0x00,mga_video_port_reg);	/* HorizontalTotal */
		mga_state.HorizontalTotal = readb(mga_video_port_val);
		writeb(0x01,mga_video_port_reg);	/* HorizDisplayEnd */
		mga_state.HorizDisplayEnd = readb(mga_video_port_val);
		writeb(0x04,mga_video_port_reg);	/* StartHorizRetrace */
		mga_state.StartHorizRetrace = readb(mga_video_port_val);
		writeb(0x05,mga_video_port_reg);	/* EndHorizRetrace */
		mga_state.EndHorizRetrace = readb(mga_video_port_val);
		writeb(0x07,mga_video_port_reg);	/* Overflow */
		mga_state.Overflow = readb(mga_video_port_val);
		writeb(0x10,mga_video_port_reg);	/* StartVertRetrace */
		mga_state.StartVertRetrace = readb(mga_video_port_val);
		writeb(0x11,mga_video_port_reg);	/* EndVertRetrace */
		mga_state.EndVertRetrace = readb(mga_video_port_val);
		writeb(0x17,mga_video_port_reg);	/* ModeControl */
		mga_state.ModeControl = readb(mga_video_port_val);
		writeb(0x01,seq_port_reg);		/* ClockingMode */
		mga_state.ClockingMode = readb(seq_port_val);
	}

	/* assure that video is enabled */
	/* "0x20" is VIDEO_ENABLE_bit in register 01 of sequencer */
	spin_lock_irq(&vga_lock);
	writeb(0x01,seq_port_reg);
	writeb(mga_state.ClockingMode | 0x20,seq_port_val);

	/* test for vertical retrace in process.... */
	if ((mga_state.CrtMiscIO & 0x80) == 0x80)
		writeb(mga_state.CrtMiscIO & 0xef,video_misc_wr);

	/*
	 * Set <End of vertical retrace> to minimum (0) and
	 * <Start of vertical Retrace> to maximum (incl. overflow)
	 * Result: turn off vertical sync (VSync) pulse.
	 */
	if (mode & VESA_VSYNC_SUSPEND) {
		writeb(0x10,mga_video_port_reg);	/* StartVertRetrace */
		writeb(0xff,mga_video_port_val); 	/* maximum value */
		writeb(0x11,mga_video_port_reg);	/* EndVertRetrace */
		writeb(0x40,mga_video_port_val);	/* minimum (bits 0..3)  */
		writeb(0x07,mga_video_port_reg);	/* Overflow */
		writeb(mga_state.Overflow | 0x84,mga_video_port_val); /* bits 9,10 of vert. retrace */
	}

	if (mode & VESA_HSYNC_SUSPEND) {
		/*
		 * Set <End of horizontal retrace> to minimum (0) and
		 *  <Start of horizontal Retrace> to maximum
		 * Result: turn off horizontal sync (HSync) pulse.
		 */
		writeb(0x04,mga_video_port_reg);	/* StartHorizRetrace */
		writeb(0xff,mga_video_port_val);	/* maximum */
		writeb(0x05,mga_video_port_reg);	/* EndHorizRetrace */
		writeb(0x00,mga_video_port_val);	/* minimum (0) */
	}

	/* restore both index registers */
	writeb(mga_state.SeqCtrlIndex,seq_port_reg);
	writeb(mga_state.CrtCtrlIndex,mga_video_port_reg);
	spin_unlock_irq(&vga_lock);	
}

static void mga_vesa_unblank(void)
{
	/* restore original values of VGA controller registers */
	spin_lock_irq(&vga_lock);
	writeb(mga_state.CrtMiscIO,video_misc_wr);

	writeb(0x00,mga_video_port_reg);		/* HorizontalTotal */
	writeb(mga_state.HorizontalTotal,mga_video_port_val);
	writeb(0x01,mga_video_port_reg);		/* HorizDisplayEnd */
	writeb(mga_state.HorizDisplayEnd,mga_video_port_val);
	writeb(0x04,mga_video_port_reg);		/* StartHorizRetrace */
	writeb(mga_state.StartHorizRetrace,mga_video_port_val);
	writeb(0x05,mga_video_port_reg);		/* EndHorizRetrace */
	writeb(mga_state.EndHorizRetrace,mga_video_port_val);
	writeb(0x07,mga_video_port_reg);		/* Overflow */
	writeb(mga_state.Overflow,mga_video_port_val);
	writeb(0x10,mga_video_port_reg);		/* StartVertRetrace */
	writeb(mga_state.StartVertRetrace,mga_video_port_val);
	writeb(0x11,mga_video_port_reg);		/* EndVertRetrace */
	writeb(mga_state.EndVertRetrace,mga_video_port_val);
	writeb(0x17,mga_video_port_reg);		/* ModeControl */
	writeb(mga_state.ModeControl,mga_video_port_val);
	writeb(0x01,seq_port_reg);		/* ClockingMode */
	writeb(mga_state.ClockingMode,seq_port_val);

	/* restore index/control registers */
	writeb(mga_state.SeqCtrlIndex,seq_port_reg);
	writeb(mga_state.CrtCtrlIndex,mga_video_port_reg);
	spin_unlock_irq(&vga_lock);
}

static void mga_pal_blank(void)
{
	int i;

	for (i=0; i<16; i++) {
		writeb (i, dac_reg) ;
		writeb (0, dac_val) ;
		writeb (0, dac_val) ;
		writeb (0, dac_val) ;
	}
}

static int mgacon_blank(struct vc_data *c, int blank, int mode_switch)
{
	switch (blank) {
	case 0:				/* Unblank */
		if (mga_vesa_blanked) {
			mga_vesa_unblank();
			mga_vesa_blanked = 0;
		}
		if (mga_palette_blanked) {
			mga_set_palette(c, color_table);
			mga_palette_blanked = 0;
			return 0;
		}
		mga_is_gfx = 0;
		/* Tell console.c that it has to restore the screen itself */
		return 1;
	case 1:				/* Normal blanking */
            	mga_pal_blank();
                mga_palette_blanked = 1;
		mgacon_set_origin(c);
		scr_memsetw((void *)mga_vram_base, BLANK, c->vc_screenbuf_size);
		return 1;
	case -1:			/* Entering graphic mode */
		scr_memsetw((void *)mga_vram_base, BLANK, c->vc_screenbuf_size);
		mga_is_gfx = 1;
		return 1;
	default:			/* VESA blanking */
            	mga_vesa_blank(blank-1);
                mga_vesa_blanked = blank;
		return 0;
	}
}

/*
 * PIO_FONT support.
 *
 * The font loading code goes back to the codepage package by
 * Joel Hoffman (joel@wam.umd.edu). (He reports that the original
 * reference is: "From: p. 307 of _Programmer's Guide to PC & PS/2
 * Video Systems_ by Richard Wilton. 1987.  Microsoft Press".)
 *
 * Change for certain monochrome monitors by Yury Shevchuck
 * (sizif@botik.yaroslavl.su).
 */

#define cmapsz 8192

static int
mgacon_do_font_op(char *arg, int set, int ch512)
{
	int i;
	char *charmap;
	int beg;
//	char * video_port_status = mga_video_port_reg + 6;
	int font_select = 0x00;

        charmap = mga_vram_base+3;
        beg = 0x06;
	
#ifdef BROKEN_GRAPHICS_PROGRAMS
	/*
	 * All fonts are loaded in slot 0 (0:1 for 512 ch)
	 */

	if (!arg)
		return -EINVAL;		/* Return to default font not supported */

	mga_font_is_default = 0;
	font_select = ch512 ? 0x04 : 0x00;
#else	
	/*
	 * The default font is kept in slot 0 and is never touched.
	 * A custom font is loaded in slot 2 (256 ch) or 2:3 (512 ch)
	 */

	if (set) {
		mga_font_is_default = !arg;
		if (!arg)
			ch512 = 0;		/* Default font is always 256 */
		font_select = arg ? (ch512 ? 0x0e : 0x0a) : 0x00;
	}

	if ( !mga_font_is_default )
		charmap += 4*cmapsz * 16;
#endif

	if (arg) {
		if (set)
			for (i=0; i<cmapsz ; i++)
				writeb(arg[i], charmap + i * 16);
		else
			for (i=0; i<cmapsz ; i++)
				arg[i] = readb(charmap + i * 16);

		/*
		 * In 512-character mode, the character map is not contiguous if
		 * we want to remain EGA compatible -- which we do
		 */

		if (ch512) {
			charmap += 2*cmapsz * 16;
			arg += cmapsz;
			if (set)
				for (i=0; i<cmapsz ; i++)
					writeb(arg[i], charmap + i * 16);
			else
				for (i=0; i<cmapsz ; i++)
					arg[i] = readb(charmap + i * 16);
		}
	}
	
	spin_lock_irq(&vga_lock);	
	writeb( 0x00, seq_port_reg );   /* First, the sequencer */
	writeb( 0x01, seq_port_val );   /* Synchronous reset */
	if (set) {
		writeb( 0x03, seq_port_reg ); /* Character Map Select */
		writeb( font_select, seq_port_val );
	}
	writeb( 0x00, seq_port_reg );
	writeb( 0x03, seq_port_val );   /* clear synchronous reset */

	spin_unlock_irq(&vga_lock);

	return 0;
}

/*
 * Adjust the screen to fit a font of a certain height
 */
static int
mgacon_adjust_height(unsigned fontheight)
{
	int rows, maxscan;
	unsigned char ovr, vde, fsr;

	if (fontheight == mga_video_font_height)
		return 0;

	mga_video_font_height = fontheight;

	rows = mga_scan_lines/fontheight;	/* Number of video rows we end up with */
	maxscan = rows*fontheight - 1;		/* Scan lines to actually display-1 */

	/* Reprogram the CRTC for the new font size */

	spin_lock_irq(&vga_lock);
	writeb( 0x07, mga_video_port_reg );		/* CRTC overflow register */
	ovr = readb(mga_video_port_val);
	writeb( 0x09, mga_video_port_reg );		/* Font size register */
	fsr = readb(mga_video_port_val);
	spin_unlock_irq(&vga_lock);

	vde = maxscan & 0xff;			/* Vertical display end reg */
	ovr = (ovr & 0xbd) +			/* Overflow register */
	      ((maxscan & 0x100) >> 7) +
	      ((maxscan & 0x200) >> 3);
	fsr = (fsr & 0xe0) + (fontheight-1);    /*  Font size register */

	spin_lock_irq(&vga_lock);
	writeb( 0x07, mga_video_port_reg );		/* CRTC overflow register */
	writeb( ovr, mga_video_port_val );
	writeb( 0x09, mga_video_port_reg );		/* Font size */
	writeb( fsr, mga_video_port_val );
	writeb( 0x12, mga_video_port_reg );		/* Vertical display limit */
	writeb( vde, mga_video_port_val );
	spin_unlock_irq(&vga_lock);

//	vc_resize_all(rows, 0);			/* Adjust console size */
	return 0;
}

static int mgacon_font_set(struct vc_data *c, struct console_font *font, unsigned flags)
{
	unsigned int charcount = font->charcount;
	int rc;
	
	if (font->width != 8 || (charcount != 256 && charcount != 512))
		return -EINVAL;
	
	rc = mgacon_do_font_op(font->data, 1, charcount == 512);
	if (rc)
		return rc;
	
	if (!(flags & KD_FONT_FLAG_DONT_RECALC))
		rc = mgacon_adjust_height(font->height);
	return rc;
}


static int mgacon_set_origin(struct vc_data *c)
{
	if (mga_is_gfx 	/* We don't play origin tricks in graphic modes */
#if 0
	    || (console_blanked && !mga_palette_blanked)
#endif
            )	/* Nor we write to blanked screens */
		return 0;
	c->vc_origin = c->vc_visible_origin = (int)video_buf;
	mga_set_mem_top(c);
	mga_rolled_over = 0;
	return 1;
}

static int mgacon_scrolldelta(struct vc_data *c, int lines)
{
    return 0;
}

static int mgacon_scroll(struct vc_data *c, int t, int b, int dir, int lines)
{
	DBG2( "_scroll %i %i %i %i\n", t, b, dir, lines);
	if (!lines)
		return 0;

	if (lines > c->vc_rows)   /* maximum realistic size */
		lines = c->vc_rows;

	switch (dir) {

	case SM_UP:
			mgacon_bmove(c,t+lines,0,t,0,b-t-lines,mga_video_num_columns);
            mgacon_clear(c,b-lines,0,lines,mga_video_num_columns);
		break;

	case SM_DOWN:
            mgacon_bmove(c,t,0,t+lines,0,b-t-lines,mga_video_num_columns);
			mgacon_clear(c,t,0,lines,mga_video_num_columns);
		break;
	}

	return 0;
}
/*
 *  The console `switch' structure for the VGA based console
 */
#define NVVGA_ADDR(x,y)  ((u16 *) mga_vram_base + (((y)*mga_video_num_columns + (x))*8))

static void mgacon_putc(struct vc_data *c, int ch, int y, int x)
{
	scr_writew(ch, NVVGA_ADDR(x, y));
}

static void mgacon_putcs(struct vc_data *c, const unsigned short *s,
		         int count, int y, int x)
{
	u16 *dest = NVVGA_ADDR(x, y);

	DBG2( "_putcs %i %i %i\n", count, y, x);
	
	for (; count > 0; count--) {
		scr_writew(*(s++), dest);
                dest+=8;
	}
}

static void mgacon_clear(struct vc_data *c, int y, int x, 
			  int height, int width)
{
	u16 *dest = NVVGA_ADDR(x, y);
	u16 eattr = c->vc_video_erase_char;

	DBG2( "_clear %i %i %i %i\n", y, x, height, width);
	if (width <= 0 || height <= 0)
		return;

        for (; height > 0; height--, dest+=mga_video_num_columns*8) {
#if 1
            scr_memsetw(dest,eattr,width*2);
#else
            int i;
            for(i=0; i<width; i++)
                scr_writew(eattr, dest+i*8);
#endif
        }
}

static void mgacon_bmove(struct vc_data *c, int sy, int sx, 
			 int dy, int dx, int height, int width)
{
	u16 *src, *dest;
    int i;

	DBG2( "_bmove %i %i %i %i %i %i\n", sy, sx, dy, dx, height, width);
	
	if (width <= 0 || height <= 0) {
		return;
		
	} else if (dy < sy || (dy == sy && dx < sx)) {
		src  = NVVGA_ADDR(sx, sy);
		dest = NVVGA_ADDR(dx, dy);

		for (; height > 0; height--) {
			for(i=0;i<width;i++)scr_writew(scr_readw(src+8*i),dest+8*i);
			src  += mga_video_num_columns*8;
			dest += mga_video_num_columns*8;
		}
	} else {
		src  = NVVGA_ADDR(sx, sy+height-1);
		dest = NVVGA_ADDR(dx, dy+height-1);

		for (; height > 0; height--) {
			for(i=width-1;i>=0;i--)scr_writew(scr_readw(src+8*i),dest+8*i);
			src  -= mga_video_num_columns*8;
			dest -= mga_video_num_columns*8;

		}
	}
}

static u16 *mgacon_screen_pos( struct vc_data *c, int offset) {
//printk(KERN_DEBUG "mgacon_screen_pos %i\n", offset);
    return (u16 *)c->vc_origin+offset;
}

struct consw mga_con = {
	.con_startup = mgacon_startup,
	.con_init = mgacon_init,
	.con_deinit = mgacon_deinit,
	.con_clear = mgacon_clear,			/* con_clear */
	.con_putc = mgacon_putc,			/* con_putc */
	.con_putcs = mgacon_putcs,			/* con_putcs */
	.con_cursor = mgacon_cursor,
	.con_scroll = mgacon_scroll,		/* con_scroll */
	.con_bmove = mgacon_bmove,			/* con_bmove */
	.con_switch = mgacon_switch,
	.con_blank = mgacon_blank,
	.con_font_set = mgacon_font_set,
	.con_set_palette = mgacon_set_palette,
	.con_scrolldelta = mgacon_scrolldelta,
#ifdef FIX_SCROLL
	.con_set_origin = mgacon_set_origin,
	.con_save_screen = mgacon_save_screen,
#else
	.con_set_origin =         NULL,
	.con_save_screen =        NULL,
#endif
	.con_build_attr = mgacon_build_attr,
	.con_invert_region = mgacon_invert_region,
	.con_screen_pos = mgacon_screen_pos,
};

#ifndef MODULE
__initfunc(void mgacon_setup(char *str, int *ints))
{

	if (ints[0] < 2)
		return;

	if (ints[1] < 1 || ints[1] > MAX_NR_CONSOLES || 
	    ints[2] < 1 || ints[2] > MAX_NR_CONSOLES)
		return;

	first_vc = ints[1]-1;
	last_vc  = ints[2]-1;
}
#endif

#ifdef MODULE
void mga_console_init(void)
#else
__initfunc(void mga_console_init(void))
#endif
{
    struct pci_dev *dev;

	video_buf=NULL;
	        
	if (first_vc > last_vc)
		return;

	dev=pci_find_device(PCI_VENDOR_ID_MATROX, PCI_ANY_ID, NULL);
	
	if(dev==NULL) {
		printk(KERN_WARNING "No Matrox card found.\n");
		return;
	}

	bar0=pci_resource_start(dev, 0);
	bar1=pci_resource_start(dev, 1);

	printk(KERN_DEBUG "MGACON: Found Matrox card, dev_id=%04x,  iomem=0x%08lx,  vidmem=0x%08lx.\n",
			dev->device, bar0, bar1);
	
	if (first_vc > last_vc)
		return;

	take_over_console(&mga_con, first_vc-1, last_vc-1, 0);
}


#ifdef MODULE

int init_module(void)
{
	mga_console_init();

	return 0;
}

void cleanup_module(void)
{
	give_up_console(&mga_con);
    
	if(video_buf!=NULL) {
		kfree(video_buf);
		iounmap(mga_vram_base);
		iounmap(mga_mmio_base0);
	}
		
}

#endif

MODULE_LICENSE("GPL");
